/*++
  Copyright (c) 2017 Microsoft Corporation

  Module Name:

  <name>

  Abstract:

  <abstract>

  Author:
  Nikolaj Bjorner (nbjorner)
  Lev Nachmanson (levnach)

  Revision History:


  --*/
#pragma once
#include "math/lp/nex.h"
namespace nla {

struct occ {
    unsigned m_occs; // number of occurences
    unsigned m_power; // min power in occurences
    occ() : m_occs(0), m_power(0) {}
    occ(unsigned k, unsigned p) : m_occs(k), m_power(p) {}
    // use the "name injection rule here"
    friend std::ostream& operator<<(std::ostream& out, const occ& c) {
        out << "(occs:" << c.m_occs <<", pow:" << c.m_power << ")";
        return out;
    }
};


// the purpose of this class is to create nex objects, keep them, and delete them

class nex_creator {

    ptr_vector<nex>                              m_allocated;
    std::unordered_map<lpvar, occ>               m_occurences_map;
    std::unordered_map<lpvar, unsigned>          m_powers;
    // the "less than" operator on expressions
    nex_lt m_lt; 
public:
    nex * clone(const nex* a) {
        switch (a->type()) {
        case expr_type::VAR: {
            auto v = to_var(a);
            return mk_var(v->var());
        }
            
        case expr_type::SCALAR: {
            auto v = to_scalar(a);
            return mk_scalar(v->value());
        }
        case expr_type::MUL: {
            auto m = to_mul(a);
            auto r = mk_mul();
            for (const auto& p : m->children()) {
                r->add_child_in_power(clone(p.e()), p.pow());
            }
            return r;
        }
        case expr_type::SUM: {
            auto m = to_sum(a);
            auto r = mk_sum();
            for (nex * e : m->children()) {
                r->add_child(clone(e));
            }
            return r;
        }
        default:
            UNREACHABLE();
            break;
        }
        return nullptr;
    }
    nex_creator(nex_lt lt) : m_lt(lt) {}
    const std::unordered_map<lpvar, occ>& occurences_map() const { return m_occurences_map; }
    std::unordered_map<lpvar, occ>& occurences_map() { return m_occurences_map; }
    const std::unordered_map<lpvar, unsigned> & powers() const { return m_powers; }
    std::unordered_map<lpvar, unsigned> & powers() { return m_powers; }
    
    void add_to_allocated(nex* r) { m_allocated.push_back(r); }

    void pop(unsigned sz) {
        for (unsigned j = sz; j < m_allocated.size(); j ++)
            delete m_allocated[j];
        m_allocated.resize(sz);
    }

    void clear() {
        for (auto e: m_allocated)
            delete e;
        m_allocated.clear();
    }
    
    unsigned size() const { return m_allocated.size(); }
    
    nex_sum* mk_sum() {
        auto r = new nex_sum();
        add_to_allocated(r);
        return r;
    }
    template <typename T>
    void add_children(T) { }
    
    template <typename T, typename K, typename ...Args>
    void add_children(T r, K e, Args ...  es) {
        r->add_child(e);
        add_children(r, es ...);
    }

    nex_sum* mk_sum(const ptr_vector<nex>& v) {
        auto r = new nex_sum();
        add_to_allocated(r);
        r->children() = v;
        return r;
    }

    nex_mul* mk_mul(const vector<nex_pow>& v) {
        auto r = new nex_mul();
        add_to_allocated(r);
        r->children() = v;
        return r;
    }

    
    template <typename K, typename...Args>
    nex_sum* mk_sum(K e, Args... es) {
        auto r = new nex_sum();
        add_to_allocated(r);
        r->add_child(e);
        add_children(r, es...);
        return r;
    }
    nex_var* mk_var(lpvar j) {
        auto r = new nex_var(j);
        add_to_allocated(r);
        return r;
    }
    
    nex_mul* mk_mul() {
        auto r = new nex_mul();
        add_to_allocated(r);
        return r;
    }

    template <typename K, typename...Args>
    nex_mul* mk_mul(K e, Args... es) {
        auto r = new nex_mul();
        add_to_allocated(r);
        add_children(r, e, es...);
        return r;
    }
    
    nex_scalar* mk_scalar(const rational& v) {
        auto r = new nex_scalar(v);
        add_to_allocated(r);
        return r;
    }

    nex * mk_div(const nex* a, lpvar j) {
        SASSERT(a->is_simplified(m_lt));
        TRACE("nla_cn_details", tout << "a=" << *a << ", v" << j << "\n";);
        SASSERT((a->is_mul() && a->contains(j)) || (a->is_var() && to_var(a)->var() == j));
        if (a->is_var())
            return mk_scalar(rational(1));
        vector<nex_pow> bv; 
        bool seenj = false;
        for (auto& p : to_mul(a)->children()) {
            const nex * c = p.e();
            int pow = p.pow();
            if (!seenj) {
                if (c->contains(j)) {
                    if (!c->is_var()) {                    
                        bv.push_back(nex_pow(mk_div(c, j)));
                        if (pow != 1) {
                            bv.push_back(nex_pow(clone(c), pow)); 
                        }
                    } else {
                        SASSERT(to_var(c)->var() == j);
                        if (p.pow() != 1) {
                            bv.push_back(nex_pow(mk_var(j), pow - 1));
                        }
                    }
                    seenj = true;
                } 
            } else {
                bv.push_back(nex_pow(clone(c)));
            }
        }
        if (bv.size() > 1) { 
            return mk_mul(bv);
        }
        if (bv.size() == 1 && bv.begin()->pow() == 1) {
            return bv.begin()->e();
        }
        if (bv.size() == 0)
            return mk_scalar(rational(1));
        return mk_mul(bv);
    }

    nex * mk_div(const nex* a, const nex* b) {
        TRACE("nla_cn_details", tout <<"("  << *a << ") / (" << *b << ")\n";);
        if (b->is_var()) {
            return mk_div(a, to_var(b)->var());
        }
        SASSERT(b->is_mul());
        const nex_mul *bm = to_mul(b);
        if (a->is_sum()) {
            nex_sum * r = mk_sum();
            const nex_sum * m = to_sum(a);
            for (auto e : m->children()) {
                r->add_child(mk_div(e, bm));
            }
            TRACE("nla_cn_details", tout << *r << "\n";);
            return r;
        }
        if (a->is_var() || (a->is_mul() && to_mul(a)->children().size() == 1)) {
            return mk_scalar(rational(1));
        }
        SASSERT(a->is_mul());
        
        const nex_mul* am = to_mul(a);
        bm->get_powers_from_mul(m_powers);
        TRACE("nla_cn_details", print_vector(m_powers, tout););
        nex_mul* ret = new nex_mul();
        for (const nex_pow& p : am->children()) {
            const nex *e = p.e();
            TRACE("nla_cn_details", tout << "e=" << *e << "\n";);
            
            if (!e->is_var()) {
                SASSERT(e->is_scalar());
                ret->add_child(mk_scalar(to_scalar(e)->value()));
                TRACE("nla_cn_details", tout << "continue\n";);
                continue;
            }
            SASSERT(e->is_var());
            lpvar j = to_var(e)->var();
            auto it = m_powers.find(j);
            if (it == m_powers.end()) {
                ret->add_child(mk_var(j));
            } else {
                it->second --;
                if (it->second == 0)
                    m_powers.erase(it);
            }
            TRACE("nla_cn_details", tout << *ret << "\n";);            
        }
        SASSERT(m_powers.size() == 0);
        if (ret->children().size() == 0) {
            delete ret;
            TRACE("nla_cn_details", tout << "return 1\n";);
            return mk_scalar(rational(1));
        }
        add_to_allocated(ret);
        TRACE("nla_cn_details", tout << *ret << "\n";);        
        return ret;
    }

};
}
